from shutil import rmtree
from typing import Type, Union, Optional

from silberstral.silberstral import create_linked_type_var

from elias.folder.run import RunFolder, _RunManagerType
from elias.manager.model import ModelManager

_ModelManagerType = create_linked_type_var(_RunManagerType, bound=ModelManager)


class ModelFolder(RunFolder[_ModelManagerType]):
    _cls_run_manager: Type[_ModelManagerType]
    _prefix: str

    def __init__(self, models_folder: str, prefix: str, localize_via_run_name: bool = False):
        self._prefix = prefix
        super(ModelFolder, self).__init__(models_folder, f"{prefix}-$[_*]", localize_via_run_name=localize_via_run_name)

    def get_prefix(self) -> str:
        return self._prefix

    def open_run(self, run_name_or_id: Union[str, int]) -> _ModelManagerType:
        run_name = self.resolve_run_name(run_name_or_id)
        return self._cls_run_manager.from_location(self._folder.get_location(),
                                                   run_name,
                                                   localize_via_run_name=self._localize_via_run_name)

    def new_run(self, name: Optional[str] = None) -> _ModelManagerType:
        new_run_name = self.generate_run_name(name)
        return self.open_run(new_run_name)

    def delete_run(self, run_name_or_id: Union[str, int]):
        run_name = self.resolve_run_name(run_name_or_id)
        rmtree(f"{self._folder.get_location()}/{run_name}")



"""
The io module contains various utility methods for loading/saving Python objects/dicts in various formats.
Currently, there is support for the most frequent data storing formats:
    - JSON (+gzip)
    - pickled Python objects (+gzip)
"""

import gzip
import io
import json
import pickle
from io import BytesIO
from pathlib import Path
from typing import Union, Tuple, Optional

from matplotlib import pyplot as plt

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import PIL.Image
import cv2
import imageio
import numpy as np
import requests
import yaml
from PIL import Image

from elias.util.fs import ensure_file_ending, ensure_directory_exists_for_file

PathType = Union[str, Path]


# =========================================================================
# Zipped JSON (.json.gz)
# =========================================================================

def save_zipped_json(obj: dict, path: PathType, suffix: str = 'json.gz'):
    """
    Parses the given Python dict into json, zips it and stores it at the specified location.
    Per default, the path will have a suffix 'json.gz'.

    Parameters
    ----------
        obj: dict
            The JSON (represented as python dict) to be zipped and stored
        path: str
            Where to store the file
        suffix: str, default 'json.gz'
            File name suffix
    """

    path = ensure_file_ending(path, suffix)
    ensure_directory_exists_for_file(path)
    with gzip.open(path, 'wb') as f:
        json.dump(obj, f)


def load_zipped_json(path: PathType, suffix: str = 'json.gz') -> dict:
    """
    Loads the specified compressed JSON file from `path` and extracts it.
    Per default, the path name is assumed to have a suffix 'json.gz'.

    Parameters
    ----------
        path: str
            Path to the zipped JSON to load
        suffix: str, default 'json.gz'
            File name suffix

    Returns
    -------
        The contents of the zipped JSON as a Python dict
    """

    path = ensure_file_ending(path, suffix)
    with gzip.open(path, 'rb') as f:
        return json.load(f)


# =========================================================================
# Zipped pickles (.p.gz)
# =========================================================================

def save_zipped_object(obj: object, path: PathType, suffix: str = 'p.gz'):
    """
    Pickles, zips and stores an arbitrary python object at the specified `path`.
    Per default, the file will have a suffix 'json.gz'.

    Parameters
    ----------
        obj: object
            The Python object to be zipped and stored
        path:
            Where to store the file
        suffix: str, default 'p.gz'
            File name suffix
    """

    path = ensure_file_ending(path, suffix)
    ensure_directory_exists_for_file(path)
    with gzip.open(path, 'wb') as f:
        pickle.dump(obj, f)


def load_zipped_object(path: PathType, suffix: str = 'p.gz') -> object:
    """
    Loads zipped and pickled Python object stored at `path`.
    Per default, the file name is assumed to have a suffix 'p.gz'.

    Parameters
    ----------
        path: str
            Path to the zipped Python object that should be loaded
        suffix: str, default 'p.gz'
            File name suffix

    Returns
    -------
        The un-pickled Python object
    """

    path = ensure_file_ending(path, suffix)
    with gzip.open(path, 'rb') as f:
        return pickle.load(f)


# =========================================================================
# Pickled objects (.p)
# =========================================================================

def save_pickled(obj: object, path: PathType, suffix: str = 'p'):
    """
    Pickles and stores an arbitrary python object at the specified `path`.
    Per default, the file will have a suffix 'p'.

    Parameters
    ----------
        obj: object
            The Python object to be pickled and stored
        path: str
            Where to store the file
        suffix: str, default 'p'
            File name suffix
    """

    path = ensure_file_ending(path, suffix)
    ensure_directory_exists_for_file(path)
    with open(f"{path}", 'wb') as f:
        pickle.dump(obj, f)


def load_pickled(path: PathType, suffix: str = 'p') -> object:
    """
    Loads a pickled Python object stored at the specified `path`.
    Per default, the file name is assumed to have a suffix 'p'.

    Parameters
    ----------
        path: str
            Path to the pickled Python object that should be loaded
        suffix: str, default 'p'
            File name suffix

    Returns
    -------
        The un-pickled Python object
    """

    path = ensure_file_ending(path, suffix)
    with open(path, 'rb') as f:
        return pickle.load(f)


# =========================================================================
# JSON (.json)
# =========================================================================

class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def save_json(obj: dict, path: PathType, suffix: str = 'json'):
    """
    Stores the given Python dict in JSON format at `path`.
    Per default, the file will have a suffix 'json'.

    Parameters
    ----------
        obj: dict
            The Python dict to store
        path: str
            Where to store the file
        suffix:
            File name suffix
    """

    path = ensure_file_ending(path, suffix)
    ensure_directory_exists_for_file(path)
    with open(path, 'w') as f:
        json.dump(obj, f, indent=4, cls=NumpyEncoder)


def load_json(path: PathType, suffix: str = 'json') -> dict:
    """
    Loads and parses the given JSON file and returns it as a Python dict.
    Per default, the file name is assumed to have a suffix 'json'.

    Parameters
    ----------
        path: str
            Path to the JSON file
        suffix: str, default 'json'
            File name suffix

    Returns
    -------
        Contents of the JSON file as Python dict
    """

    path = ensure_file_ending(path, suffix)
    with open(path, 'r') as f:
        return json.load(f)


# =========================================================================
# YAML (.yaml)
# =========================================================================

def save_yaml(obj: dict, path: PathType, suffix: str = 'yaml'):
    path = ensure_file_ending(path, suffix)
    ensure_directory_exists_for_file(path)
    with open(path, 'w') as f:
        yaml.dump(obj, f)


def load_yaml(path: PathType, suffix: str = 'yaml') -> dict:
    path_with_suffix = ensure_file_ending(path, suffix)
    if not Path(path_with_suffix).exists():
        if suffix == 'yaml':
            # Try loading .yml instead
            return load_yaml(path, 'yml')
        else:
            raise FileNotFoundError(f"Could not load YAML file {path_with_suffix} (Neither with .yaml suffix). "
                                    f"Is the path correct?")

    with open(path_with_suffix, 'r') as f:
        return yaml.load(f, Loader=yaml.FullLoader)


# =========================================================================
# Images
# =========================================================================

def save_img(img: np.ndarray, path: PathType):
    """
    Save the given numpy array as an 8-bit image. The image type is determined by the file extension.
    `image` should be a np.uint8 array with shape [H, W, (C)] with C in {1, 3, 4} and values in range [0, 255]
    If `image` has type np.float32, it is assumed that values are in range [0, 1]. The float image will be automatically converted to uint8 type.

    Parameters
    ----------
        img: image as a [H, W, (C)] numpy array with values in [0, 255] or [0, 1].
        path: where to store the image. Any necessarily folders will be created automatically.
    """

    ensure_directory_exists_for_file(path)

    if Path(path).suffix == '.exr':
        imageio.imwrite(path, img)
    else:
        if img.dtype == np.float32:
            # TODO: Do we really want to quantize without asking?
            # If float array was passed, have to transform [0.0, 1.0] -> [0, ..., 255)
            assert 0 <= img.min() and img.max() <= 1, \
                "passed float array should have values between 0 and 1 to be interpreted as image"
            img = (img * 255).astype(np.uint8)

        if len(img.shape) == 3 and img.shape[2] == 1:
            # Assume this should be stored as a single-channel grayscale image
            img = img[:, :, 0]

        Image.fromarray(img).save(path)


def load_img(path: PathType) -> np.ndarray:
    if Path(path).suffix == '.exr':
        img = imageio.imread_v2(path)
    else:
        img = Image.open(path)

    return np.asarray(img)


def download_img(url: str) -> np.ndarray:
    response = requests.get(url)
    pil_img = Image.open(BytesIO(response.content))
    img = np.asarray(pil_img)

    return img


InterpolationType = Literal['nearest', 'bilinear', 'bicubic', 'lanczos']


def resize_img(img: np.ndarray,
               scale: Union[float, Tuple[float, float]],
               interpolation: InterpolationType = 'bilinear',
               use_opencv: bool = False) -> np.ndarray:
    try:
        iter(scale)
    except TypeError:
        scale_x = scale
        scale_y = scale
    else:
        scale_x, scale_y = scale

    if use_opencv:
        if interpolation == 'nearest':
            interpolation = cv2.INTER_NEAREST
        elif interpolation == 'bilinear':
            interpolation = cv2.INTER_LINEAR
        elif interpolation == 'bicubic':
            interpolation = cv2.INTER_CUBIC
        elif interpolation == 'lanczos':
            interpolation = cv2.INTER_LANCZOS4
        else:
            raise ValueError(f"Invalid interpolation type: {interpolation}")

        img = cv2.resize(img, (int(img.shape[1] * scale_x), int(img.shape[0] * scale_y)), interpolation=interpolation)
    else:
        if interpolation == 'nearest':
            interpolation = PIL.Image.Resampling.NEAREST
        elif interpolation == 'bilinear':
            interpolation = PIL.Image.Resampling.BILINEAR
        elif interpolation == 'bicubic':
            interpolation = PIL.Image.Resampling.BICUBIC
        elif interpolation == 'lanczos':
            interpolation = PIL.Image.Resampling.LANCZOS
        else:
            raise ValueError(f"Invalid interpolation type: {interpolation}")

        img_pil = Image.fromarray(img)
        img_pil = img_pil.resize((int(img.shape[1] * scale_x), int(img.shape[0] * scale_y)), resample=interpolation)
        img = np.array(img_pil)

    return img


def fig2img(fig: Optional[plt.Figure] = None, transparent: bool = False) -> np.ndarray:
    """
    Converts a Matplotlib figure to a numpy image.
    If no figure is given, the current open matplot figure will be used instead.
    """

    buf = io.BytesIO()
    if fig is None:
        fig = plt.gcf()

    fig.savefig(buf, bbox_inches='tight', transparent=transparent)
    buf.seek(0)
    img = np.asarray(Image.open(buf))

    if transparent:
        return img[..., :4]
    else:
        return img[..., :3]



from importlib import import_module
from typing import Type, List


def ensure_type(obj, cls: Type):
    """
    Ensures that `obj` is of type `cls` by calling `cls`'s constructor in case it isn't.
    """
    if not isinstance(obj, cls):
        return cls(obj)
    else:
        return obj


def ensure_list(obj) -> List:
    if not isinstance(obj, list):
        return [obj]
    else:
        return obj


def class_to_module_path(cls: Type) -> str:
    module = cls.__module__
    if module == 'builtins':
        module_path = cls.__qualname__  # avoid outputs like 'builtins.str'
    else:
        module_path = module + '.' + cls.__qualname__

    return module_path


def module_path_to_class(dotted_path: str) -> Type:
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """

    parts = dotted_path.rsplit('.', 1)
    if len(parts) == 2:
        module_path = parts[0]
        class_name = parts[1]
    else:
        module_path = 'builtins'
        class_name = parts[0]

    module = import_module(module_path)
    return getattr(module, class_name)



from enum import Enum, auto
from pathlib import Path
from typing import Callable

from elias.util.io import save_json, save_yaml, load_json, load_yaml


class ArtifactType(Enum):
    JSON = auto()
    YAML = auto()

    def get_saver(self) -> Callable[[dict, str], None]:
        if self == ArtifactType.JSON:
            return save_json
        elif self == ArtifactType.YAML:
            return save_yaml
        else:
            raise ValueError(f"Unknown Artifact type: {self}")

    def get_loader(self) -> Callable[[str], dict]:
        if self == ArtifactType.JSON:
            return load_json
        elif self == ArtifactType.YAML:
            return load_yaml
        else:
            raise ValueError(f"Unknown Artifact type: {self}")

    def get_file_ending(self) -> str:
        if self == ArtifactType.JSON:
            return 'json'
        elif self == ArtifactType.YAML:
            return 'yaml'
        else:
            raise ValueError(f"Unknown Artifact type: {self}")


class ArtifactManager:

    def __init__(self, location: str, artifact_type: ArtifactType = ArtifactType.JSON):
        assert Path(location).is_dir(), f"Specified location '{location}' is not a directory"

        self._location = location
        self._artifact_type = artifact_type

    def _save_artifact(self, artifact: dict, name: str):
        self._artifact_type.get_saver()(artifact, f"{self._location}/{name}")

    def _load_artifact(self, name: str) -> dict:
        return self._artifact_type.get_loader()(f"{self._location}/{name}")



from .analysis import Analysis
from .artifact import ArtifactManager
from .buffered import BufferedDataLoader
from .data import BaseDataManager, BaseSampleDataManager, BaseSliceDataManager, RandomAccessSampleDataManager, \
    RandomAccessSliceDataManager
from .model import ModelManager



import re
from typing import Union, TypeVar, Generic, List, Optional

from silberstral import reveal_type_var

from elias.folder.folder import Folder
from elias.util import ensure_directory_exists
from elias.util.version import Version

_DataManagerType = TypeVar("_DataManagerType", bound='BaseDataManager')

DATASET_VERSION_REGEX = re.compile(r"v(\d+(?:\.\d+)*)(?:-.*)?")


class DataFolder(Folder, Generic[_DataManagerType]):
    """
    A DataFolder refers to a file system folder that contains the artefacts created by a data preprocessing stage.
    Each preprocessing run is stored in a separate sub folder.
    Research workflows often exhibit several iterations of data preprocessing which is why the preprocessing runs are
    tagged with a version.
    Names of sub folders are assumed to follow the format `v{version}-{name}`.
    Examples: "v1-first-dataset" or "v2.3.1-imagenet-with-more-dogs".

    DataFolders should be subclassed for each specific dataset type, e.g., one for ImageNet, one for MNIST, etc...
    During subclassing _DataManagerType is replaced by the actual class of the data manager that handles this dataset
    type.
    """

    _version_levels: int
    _default_bump_level: int
    _cls_data_manager: _DataManagerType

    def __init__(self,
                 location: str,
                 version_levels: int = 2,
                 default_bump_level: int = -1,
                 localize_via_run_name: bool = False):
        """
        Parameters
        ----------
            location:
                folder in which the preprocessing runs are stored
            version_levels:
                how many levels should be used for versioning, i.e., v1.2 or v1.2.3.4 etc.
            default_bump_level:
                which level should be increased when a new dataset is created
            localize_via_run_name:
                whether only the dataset name will be used for finding the corresponding single dataset folder
        """
        ensure_directory_exists(location)
        super(DataFolder, self).__init__(location)

        self._version_levels = version_levels
        self._default_bump_level = default_bump_level
        self._cls_data_manager = reveal_type_var(self, _DataManagerType)
        self._localize_via_run_name = localize_via_run_name

    def list_datasets(self) -> List[str]:
        """
        Lists all sub folders that obey the `v{version}-{name}` format.

        Returns
        -------
            a list containing the found dataset names
        """

        dataset_versions = self.list_dataset_versions()
        dataset_names = [self.get_dataset_name_by_version(version) for version in dataset_versions]
        return dataset_names
        # dataset_folders = self.ls()
        # dataset_folders = [f for f in dataset_folders if DATASET_VERSION_REGEX.match(f)]
        # return dataset_folders

    def list_dataset_versions(self) -> List[Version]:
        """
        Lists all the versions from the sub folders that obey the `v{version}-{name}` format.

        Returns
        -------
            a list containing the found datasets' versions
        """

        dataset_folders = self.ls()
        dataset_versions = []

        for folder in dataset_folders:
            p = DATASET_VERSION_REGEX.match(folder)
            if p:
                version_specifier = p.group(1)
                dataset_versions.append(Version(version_specifier))

        return sorted(dataset_versions)

    def get_dataset_name_by_version(self, dataset_version: Union[str, Version]) -> str:
        """
        Given its version, retrieves the full dataset name of a sub folder that obeys the `v{version}-{name}` format.

        Parameters
        ----------
            dataset_version: the version to search for

        Returns
        -------
            the full name of the dataset the belongs to the specified version
        """

        if isinstance(dataset_version, str) and dataset_version.startswith('v'):
            dataset_version = dataset_version[1:]
        dataset_folders = self.ls()

        for folder in dataset_folders:
            p = DATASET_VERSION_REGEX.match(folder)
            if p:
                version_specifier = p.group(1)
                if version_specifier == dataset_version:
                    return folder

        raise ValueError(f"Could not find dataset version `{dataset_version}` in folder `{self._location}`")

    def open_dataset(self, dataset_version: Union[str, Version]) -> _DataManagerType:
        """
        Navigates into the sub folder that belongs to the specified dataset version.
        A data manager of the respective type is created rooted at that folder for easy interaction with the stored
        artifacts.

        Parameters
        ----------
            dataset_version: version or full name of the dataset that shall be opened

        Returns
        -------
            a data manager to interact with the dataset files
        """

        dataset_name = self._get_full_dataset_name(dataset_version)

        # Create a new data manager of the corresponding class for the specified dataset folder
        return self._cls_data_manager.from_location(self._location,
                                                    dataset_name,
                                                    localize_via_run_name=self._localize_via_run_name)

    def create_dataset(self, name: Optional[str] = None, bump_level: Optional[int] = None) -> _DataManagerType:
        """
        Creates a new data manager for a dataset with the given name. It is automatically assigned a version that is
        higher than the existing ones.

        Parameters
        ----------
            name: name of the new dataset (without version specifier, this will be added automatically)
            bump_level:
                which level of the version should be increased. Levels are 0-indexed.
                If `None`, the default specified in the constructor is used

        Returns
        -------
            A data manager for the newly created (empty) dataset folder
        """

        bump_level = self._default_bump_level if bump_level is None else bump_level
        if len(self.list_datasets()) == 0:
            # First dataset in this folder. Initialize with 0.0.1 (or similar)
            initial_version = Version.from_zero(self._version_levels)
            initial_version.bump(bump_level)
            new_version = initial_version
        else:
            # Some datasets already exist. Find maximum version and bump
            versions = self.list_dataset_versions()
            max_version = max(versions)
            max_version.bump(bump_level)
            new_version = max_version

        dataset_name = f"v{new_version}" if name is None else f"v{new_version}-{name}"
        self.mkdir(dataset_name)

        return self._cls_data_manager.from_location(self._location,
                                                    dataset_name,
                                                    localize_via_run_name=self._localize_via_run_name)

    def remove_dataset(self, dataset_version: Union[str, Version]):
        """
        Deletes the specified dataset folder and all its content.

        Parameters
        ----------
            dataset_version: version or full name of the dataset to be deleted
        """

        dataset_name = self._get_full_dataset_name(dataset_version)
        self.rmdir(dataset_name)

    def _get_full_dataset_name(self, dataset_version: Union[str, Version]) -> str:
        dataset_version = str(dataset_version)
        if Version.is_valid(dataset_version):
            # Only version was given, need to find corresponding dataset name
            return self.get_dataset_name_by_version(dataset_version)
        else:
            # Full dataset name (including version specifier) was given
            return dataset_version



from abc import ABC, abstractmethod
from typing import Iterable, Iterator, Union, List, Set, TypeVar

_T = TypeVar('_T')


class IterableDataLoader(ABC):
    @abstractmethod
    def __iter__(self):
        pass


class RandomAccessDataLoader(Iterable[_T]):

    def __iter__(self) -> Iterator[_T]:
        return (self[idx] for idx in range(len(self)))

    def __getitem__(self, idx: Union[int, slice]) -> Union[_T, List[_T]]:
        if isinstance(idx, slice):
            return self.get_slice(idx)
        elif isinstance(idx, int):
            return self._get_single_item(idx)
        else:
            try:
                idx_int = int(idx)
                return self._get_single_item(idx_int)
            except TypeError:
                raise ValueError(f"Unsupported index type passed to __getitem__: {type(idx)}")

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def _get_single_item(self, idx: int) -> _T:
        pass

    def get_slice(self, index_slice: slice) -> List[_T]:
        """
        Retrieves the specified slice from the dataloader. As opposed to `view()`, the items are not lazyly loaded but
        directly retrieved.

        Parameters
        ----------
            index_slice: The indices (in form of a slice) to retrieve

        Returns
        -------
            a list containing the requested items
        """

        return [self[idx] for idx in range(*index_slice.indices(len(self)))]

    def view(self,
             indices: Union[slice, List[int], Set[int]],
             exclude: bool = False) -> 'RandomAccessDataLoaderView[_T]':
        """
        Provides a proxy data manager that will only iterate over the given indices.
        If `exclude` is set the proxy data manager will access all elements except those specified by `indices`

        Parameters
        ----------
            indices: Indices to include/exclude in the proxy data manager
            exclude: whether the specified indices should be included or excluded

        Returns
        -------
            A proxy data manager
        """

        # TODO: find a way to inherit self completely and basically only change the __get_item__() method
        return RandomAccessDataLoaderView(self, indices, exclude=exclude)


class RandomAccessDataLoaderView(RandomAccessDataLoader[_T]):
    """
    Provides simple means of changing the iteration over a dataloader without copying the underlying data.
    """

    def __init__(self,
                 dataloader: RandomAccessDataLoader[_T],
                 indices: Union[slice, List[int], Set[int]],
                 exclude: bool = False):
        """
        Parameters
        ----------
            dataloader: the underlying dataloader which is viewed at
            indices: the indices of the respective elements that will be used by the dataloader view. When the view
                is iterated over, the order of passed indices will be used.
                If a set is passed, the order of indices is undefined.
        """

        self._dataloader = dataloader
        n_samples = len(self._dataloader)
        if isinstance(indices, slice):
            ranged_indices = range(n_samples)[indices]
            # TODO: DataLoader can only be iterated over once
            if exclude:
                ranged_indices = set(ranged_indices)
                self._indices = (idx for idx in range(n_samples) if idx not in ranged_indices)
            else:
                self._indices = ranged_indices
        elif isinstance(indices, list) or isinstance(indices, set):
            assert not indices or max(indices) < n_samples, \
                f"Cannot create view with index {max(indices)} for data loader with length {n_samples}"
            if exclude:
                indices = indices if isinstance(indices, set) else set(indices)
                self._indices = [idx for idx in range(n_samples) if idx not in indices]
            else:
                self._indices = list(indices)
        else:
            raise ValueError(f"view indices must be slice or list not {type(indices)}")

    def __iter__(self) -> Iterator[_T]:
        return (self._dataloader[idx] for idx in self._indices)

    def _get_single_item(self, idx: int) -> _T:
        return self._dataloader[self._indices[idx]]

    def __len__(self) -> _T:
        return len(self._indices)



from math import ceil
from typing import Iterator, Iterable, Generator, List


def batchify(generator: Iterable, batch_size: int, lazy: bool = False) -> Generator[List, None, None]:
    """
    Lazyly evaluated batch-wise loading
    """

    if batch_size == 1:
        for item in generator:
            yield item
        return

    if lazy:
        # Lazy returns batches as a generator where objects are only touched upon actually querying them
        iterator = iter(generator)
        try:
            while True:
                first = next(iterator)

                def chunk():
                    try:
                        yield first
                        for _ in range(batch_size - 1):
                            yield next(iterator)
                    except StopIteration:
                        pass

                yield chunk()
        except StopIteration:
            pass
    else:
        # Regular mode materializes all objects within a batch before the batch is returned as a list
        batch = []
        for i, item in enumerate(generator):
            batch.append(item)
            if (i + 1) % batch_size == 0:
                yield batch
                batch = []
        if batch:
            yield batch


def batchify_sliced(tensor, batch_size: int) -> Iterator:
    try:
        n_samples = len(tensor)
    except Exception:
        try:
            n_samples = tensor.shape[0]
        except Exception:
            raise ValueError(f"Cannot infer length of passed tensor with type {type(tensor)}. "
                             f"Ensure to use a common Tensor/Array format")

    n_batches = ceil(n_samples / batch_size)
    for i_batch in range(n_batches):
        if i_batch == n_batches - 1:
            yield tensor[i_batch * batch_size:]  # Return all remaining samples as the last batch
        else:
            yield tensor[i_batch * batch_size: (i_batch + 1) * batch_size]


from abc import abstractmethod
from typing import List


class CombinedIterableStopCriterion:

    @abstractmethod
    def should_stop(self, just_depleted_dl_idx: int, remaining_data_loader_indices: List[int]) -> bool:
        pass


class CombinedIterableStopCriterionAnyEmpty(CombinedIterableStopCriterion):

    def should_stop(self, just_depleted_dl_idx: int, remaining_data_loader_indices: List[int]) -> bool:
        return True


class CombinedIterableStopCriterionAllEmpty(CombinedIterableStopCriterion):

    def should_stop(self, just_depleted_dl_idx: int, remaining_data_loader_indices: List[int]) -> bool:
        return len(remaining_data_loader_indices) == 0


class CombinedIterableStopCriterionSpecificEmpty(CombinedIterableStopCriterion):

    def __init__(self, specific_dl_idx: int):
        self._specific_dl_idx = specific_dl_idx

    def should_stop(self, just_depleted_dl_idx: int, remaining_data_loader_indices: List[int]) -> bool:
        return just_depleted_dl_idx == self._specific_dl_idx



from collections import defaultdict
from statistics import mean, median
from time import time
from typing import Dict, Optional

from numpy import std
from tabulate import tabulate


class TimeMeasurer:

    def __init__(self):
        self.start = time()
        self.times = []

    def measure(self) -> float:
        now = time()
        measured_time = now - self.start
        self.times.append(measured_time)
        self.start = now
        return measured_time

    def print_measure(self, text: str):
        passed_time = self.measure()
        print(f"{text}: {passed_time*1000:.2f}ms")

    def __len__(self) -> int:
        return len(self.times)

    def __getitem__(self, i) -> float:
        return self.times[i]

    def __str__(self) -> str:
        if len(self.times) == 1:
            return f"{self.times[0]:.2f}"
        return f"{self.times}"


class Timing:
    """
    Usage:

    with Timing() as t:
        ...
    print(f"took {t[0]:0.2} seconds")

    Within the with-statement we can take several measurements by calling t.measure().
    Every measurement will be available after the closure in the list t.
    """

    def __enter__(self) -> TimeMeasurer:
        self.time_measurer = TimeMeasurer()
        return self.time_measurer

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.time_measurer.measure()


class LoopTimer:
    def __init__(self,
                 max_iterations: Optional[int] = None,
                 warmup: int = 0,
                 print_at_last_iteration: bool = False,
                 disable: bool = False):
        """
        Utility to time repeated calculations in a loop and aggregate execution times.

        Parameters
        ----------
        max_iterations:
            If set, only the first n iterations will be recorded
        warmup:
            Optionally, skip the first n iterations before timings will be recorded (Often loops are slower in the beginning until caching kicks in)
        print_at_last_iteration:
            If set, a timing summary will be printed automatically once new_iteration() is called for the first time after max_iterations has been reached

        Usage
        -----
        >>> timer = LoopTimer
        >>> for i in range(100):
        >>>     timer.new_iteration()
        >>>     cmd1()
        >>>     timer.measure("cmd1")
        >>>     cmd2()
        >>>     timer.measure("cmd2")
        >>>
        >>> timer.print_summary()
        """

        self.start = time()
        self.loop_start = None
        self.times = defaultdict(list)

        self._start_iteration = warmup
        self._end_iteration = max_iterations + warmup if max_iterations is not None else None
        self._current_iteration = 0
        self._print_at_last_iteration = print_at_last_iteration
        self._disable = disable

    def measure(self, name: str) -> Optional[float]:
        if self._disable:
            return None

        now = time()
        if self._start_iteration < self._current_iteration and (self._end_iteration is None or self._current_iteration < self._end_iteration):
            measured_time = now - self.start
            self.times[name].append(measured_time)
        else:
            measured_time = None
        self.start = now
        return measured_time

    def start(self):
        """
        Set start time for next measurement to now.
        """
        if not self._disable:
            self.start = time()

    def new_iteration(self):
        """
        Intended to be called at the beginning of a new iteration.
        """

        if not self._disable:
            self.start = time()

            # Note down how long the whole iteration took
            if (self.loop_start is not None
                    and self._start_iteration < self._current_iteration
                    and (self._end_iteration is None or self._current_iteration < self._end_iteration)):
                measured_loop_time = self.start - self.loop_start
                self.times["loop iteration"].append(measured_loop_time)

            self.loop_start = self.start
            self._current_iteration += 1

            if self._end_iteration is not None and self._current_iteration == self._end_iteration:
                self.print_summary()

    def get_iteration_time(self) -> float:
        """
        Returns
        -------
            the duration of the current loop iteration since new_iteration() in seconds
        """
        return time() - self.loop_start

    def summary(self) -> Dict[str, float]:
        if self._disable:
            return dict()

        summary = {name: mean(times) for name, times in self.times.items()}
        return summary

    def print_summary(self):
        if not self._disable:
            print("===========================")
            print("Timing Summary")
            print("===========================")
            summary = {name: [mean(times), median(times), std(times)] for name, times in self.times.items()}
            sorted_summary = sorted(summary.items(), key=lambda x: x[1], reverse=True)
            rows = []
            for name, (mean_time, median_time, std_time) in sorted_summary:
                row = [name, f"{median_time*1000:.3f}ms", f"{mean_time*1000:.3f}ms ± {std_time*1000:.3f}ms"]
                rows.append(row)

            print(tabulate(rows, headers=["section", "median", "mean+std"]))



import bisect
import re
from collections import OrderedDict
from typing import Optional

import cv2
import numpy as np
from PIL import Image

from elias.util import ensure_directory_exists_for_file


def make_wandb_video(wandb_project: str,
                     run_id: str,
                     image_key: str,
                     output_folder: str,
                     /,
                     max_n_frames: int = 30,
                     until_step: Optional[int] = None,
                     fps: int = 10):
    """
    Collects all images that were logged with a given key to a wandb run.
    Images are packed into a .mp4 video for playback.

    Parameters
    ----------
        wandb_project:
            name of the wandb project
        run_id:
            Id of the wandb run. Can be retrieved from the URL: https://wandb.ai/$USER/diff-vp/runs/plh665gq <- run ID
        image_key:
            Which subset of images that are logged to the run should be downloaded. The image key is usually written
            in bold font above the logged image in the dashboard
        output_folder:
            Path to a folder in which the .mp4 will be stored
        max_n_frames:
            Sometimes, many images are logged to a run.
            To avoid excessive downloads, a maximum number of frames is specified
        until_step:
            If specified, only images until the given step will be downloaded
        fps:
            Playback speed of the generated video
    """

    # TODO: Currently, models are imported on-demand
    #   Wanted to avoid adding all these dependencies just for one method
    #   Maybe need yet another different repository for these kind of training/logging helpers?
    import wandb
    from tqdm import tqdm
    import requests
    import mediapy

    entity = wandb.setup()._get_username()
    api = wandb.Api()
    run = api.run(f"{entity}/{wandb_project}/{run_id}")

    output_path = f"{output_folder}/{run.name}_{image_key.replace('/', '-')}.mp4"
    ensure_directory_exists_for_file(output_path)

    # Collect all image urls from the wandb run that match the specified image key
    extract_step_regex = re.compile(f"media/images/{re.escape(image_key)}_(\d+).*")
    image_files = OrderedDict()
    for file in tqdm(run.files(), desc="Collect images"):
        matched = extract_step_regex.match(file.name)
        if matched:
            step = int(matched.group(1))
            image_files[step] = file

    image_files = OrderedDict(sorted(image_files.items()))
    image_files_values = list(image_files.values())

    # Select at most max_n_frames images
    logged_steps = list(image_files.keys())
    if until_step is not None:
        logged_steps = [step for step in logged_steps if step <= until_step]
    min_step = min(logged_steps)
    max_step = max(logged_steps)

    n_steps = len(logged_steps)
    if n_steps < max_n_frames:
        # Use all logged images
        steps = logged_steps
    else:
        # Use an equally spaced selection of images
        steps = np.linspace(min_step, max_step, max_n_frames)

    # Performance improvement: In case n_steps < max_n_frames, we will use some frames multiple times
    # We can avoid downloading them unnecessarily by caching the last image
    previous_step = None
    cached_image = None

    writer = None
    for step in tqdm(steps, desc="Downloading images"):
        # Find image file with step closest to requested step
        closest_step_idx = bisect.bisect_left(logged_steps, step)
        file = image_files_values[closest_step_idx]
        closest_step = logged_steps[closest_step_idx]

        # Actually download artifact and parse into image
        if closest_step == previous_step and cached_image is not None:
            image_numpy = cached_image
        else:
            response = requests.get(file.url, auth=("api", api.api_key), stream=True, timeout=5)
            image = Image.open(response.raw)
            image_numpy = np.array(image)
            cached_image = image_numpy
        previous_step = closest_step

        # Font-face:
        # 1 -> 20px
        # 2 -> 40px
        # Good height is ~10% of image width
        font_face = image_numpy.shape[1] / 10 / 20
        text = f"{closest_step}"
        cv2.putText(image_numpy, text, (10, image_numpy.shape[0] - 10), 0, font_face, (0, 255, 0))

        # Many mp4 player want an even number of pixels for width and height
        H_even = 2 * int(image_numpy.shape[0] / 2)
        W_even = 2 * int(image_numpy.shape[1] / 2)
        image_numpy = image_numpy[:H_even, :W_even]

        # Fuse image into video and stream to disk
        if writer is None:
            render_width = int(image_numpy.shape[1])
            render_height = int(image_numpy.shape[0])
            writer = mediapy.VideoWriter(
                path=output_path,
                shape=(render_height, render_width),
                fps=fps,
            )
            writer.__enter__()
        writer.add_image(image_numpy)

    writer.__exit__()



import logging
from asyncio import Event
from queue import Queue
from threading import Thread
from typing import Iterable, Iterator, Sized, TypeVar, Optional, Type, Any

from elias.config import Config
from elias.manager.data import BaseDataManager
from elias.util.timing import Timing

_SampleType = TypeVar('_SampleType')

# A special message that is used for internal queues to signalize that the producer thread is done
_QUEUE_END_MSG = object()


class BufferedDataLoader(Iterable[_SampleType]):
    """
    Wrapper class for arbitrary data managers that preloads samples in the background and provides asynchroneous saving.
    Useful in multiprocessing settings where we can have the main process preloading data while other processes do the
    work.
    The idea is to eliminate any waiting when iterating over the samples or when saving a dataset slice. This is
    obtained by using background worker threads that operate on queues instead of directly using the data_manager.
    To ensure that the python process can end after using a BufferedDataManager, one should call the .shutdown() method
    """

    _data_loader: Iterable[_SampleType]
    _load_buffer: Queue
    _load_worker: Optional[Thread]
    _stop_event: Event

    def __init__(self, data_loader: Iterable[_SampleType], size_load_buffer: int = 5000):
        """

        Parameters
        ----------
            data_loader:
                can be any iterable that provides samples
            size_load_buffer:
                specifies how many samples will be prefetched from the `data_loader`
        """

        self._data_loader = data_loader
        self._load_buffer = Queue(size_load_buffer)
        self._load_worker = None  # Will be initialized upon obtaining an iterator
        self._stop_event = Event()

    def __iter__(self) -> Iterator[_SampleType]:
        """
        Initializes a worker for prefetching data. The worker will start populating the internal queue once an iterator
        is created. To avoid spawning multiple workers, one can only have one iterator at a time.
        """

        if self._load_worker is not None:
            raise Exception("There is already an iterator running!")
        self._load_worker = self.LoadWorker(self._data_loader, self._load_buffer, self._stop_event)
        self._load_worker.start()
        return BufferedDataLoader.Iterator(self)

    def __len__(self) -> int:
        try:
            if isinstance(self._data_loader, Sized):
                return len(self._data_loader)
        except TypeError:
            pass

        raise TypeError("Underlying dataloader did not specify len()")

    def __del__(self):
        """
        Destructor. Attempts to join all threads to allow the python script to exit cleanly.
        """

        self.shutdown()

    def shutdown(self):
        """
        Clears all the buffers, terminates all workers and prepares the buffered data manager to be used again.
        Should be called when one is done with iterating over the samples to allow the python process to end.
        """

        self._stop_event.set()  # Signalize the load worker to shutdown
        if self._load_worker:
            if self._load_worker.is_alive() and not self._load_buffer.empty():
                # In this case, the load worker is waiting to put something into the queue and thus cannot receive the
                # stop signal. Resolve by taking one element out of the read buffer
                self._load_buffer.get()
            self._load_worker.join()

        self._load_buffer.queue.clear()
        self._stop_event = Event()
        self._load_worker = None

    # -------------------------------------------------------------------------
    # Inner classes
    # -------------------------------------------------------------------------

    class Iterator(Iterator[_SampleType]):

        _buffered_data_loader: 'BufferedDataLoader'

        def __init__(self, buffered_data_loader: 'BufferedDataLoader'):
            self._buffered_data_loader = buffered_data_loader

        def __next__(self) -> _SampleType:
            """
            Reads from the internal buffer and only blocks when it is empty. In this case, it might help to increase the
            size of the internal buffer via size_load_buffer
            """

            data = self._buffered_data_loader._load_buffer.get()
            if data == _QUEUE_END_MSG:
                # the load worker will put a special DONE MESSAGE to the internal queue to signal that the data_manager
                # won't provide more samples
                self._buffered_data_loader._load_worker.join()
                self._buffered_data_loader._load_worker = None
                raise StopIteration
            return data

    class LoadWorker(Thread):
        """
        Background thread that iterates over all samples in data_manager and puts them onto the internal load buffer.
        To avoid out of memory issues, the internal queue has limited size which can be controlled via size_load_buffer
        """

        _data_loader: Iterable[_SampleType]
        _read_buffer: Queue
        _stop_event: Event

        def __init__(self, data_loader: Iterable[_SampleType], read_buffer: Queue, stop_event: Event):
            Thread.__init__(self)
            self._data_loader = data_loader
            self._read_buffer = read_buffer
            self._stop_event = stop_event

        def run(self) -> None:
            with Timing() as t:
                for sample in self._data_loader:
                    logging.debug(f"Loading sample took {t.measure(): .3f}s")

                    if self._stop_event.is_set():
                        return
                    self._read_buffer.put(sample)
                    # Signalize that the data_manager iterator is empty
                self._read_buffer.put(_QUEUE_END_MSG)


class BufferedDataManager(BaseDataManager[_SampleType, Config, Config]):
    """
    Wrapper class for arbitrary data managers that preloads samples in the background and provides asynchroneous saving.
    Useful in multiprocessing settings where we can have the main process preloading data while other processes do the
    work.
    The idea is to eliminate any waiting when iterating over the samples or when saving a dataset slice. This is
    obtained by using background worker threads that operate on queues instead of directly using the data_manager.
    To ensure that the python process can end after using a BufferedDataManager, one should call the .shutdown() method
    """

    _data_manager: BaseDataManager
    _buffered_data_loader: BufferedDataLoader
    _save_buffer: Queue
    _save_worker: Optional[Thread]
    _stop_event: Event

    def __init__(self, data_manager: BaseDataManager, size_load_buffer: int = 5000, size_save_buffer: int = 1):
        """

        Parameters
        ----------
            data_manager:
                can be an arbitrary data manager that supports iterating over samples and saving dataset files
            size_load_buffer:
                specifies how many SAMPLES will be prefetched from data_manager
            size_save_buffer:
                specifies how many DATASET SLICES will be buffered until a call to .save() will actually block
        """

        super(BufferedDataManager, self).__init__(data_manager._root_location,
                                                  data_manager._run_name,
                                                  data_manager._file_name_format,
                                                  shuffle=data_manager._shuffle,
                                                  artifact_type=data_manager._artifact_type)

        self._data_manager = data_manager
        self._buffered_data_loader = BufferedDataLoader(data_manager, size_load_buffer=size_load_buffer)
        self._save_buffer = Queue(size_save_buffer)
        self._save_worker = None  # Will be initialized when the first path needs to be saved

        # Retrieve classes for Data config and statistics from the provided data manager
        # This is the reason why _ConfigType and _StatisticsType are set to None in the inheritance
        # of BufferedDataManager. The BufferedDataManager essentially wraps and mimics the provided data manager
        self._config_cls = data_manager._config_cls
        self._statistics_cls = data_manager._statistics_cls

    @classmethod
    def from_location(cls: Type['BufferedDataManager'], location: str) -> 'BufferedDataManager':
        raise NotImplementedError("Default BufferedDataManager does not provide initialization with from_location()")

    def __iter__(self) -> Iterator[_SampleType]:
        """
        Initializes a worker for prefetching data. The worker will start populating the internal queue once an iterator
        is created. To avoid spawning multiple workers, one can only have one iterator at a time.
        """

        return iter(self._buffered_data_loader)

    def __del__(self):
        """
        Destructor. Attempts to join all threads to allow the python script to exit cleanly.
        """

        self.shutdown()

    def save(self, data, **kwargs):
        """
        Puts the data on the internal save buffer and immediately returns. Only blocks when the internal save buffer
        is already full, i.e., the worker takes too longer to save one dataset slice than new data is incoming.
        In this case, this code has to be extended to allow for multiple save workers.
        A save worker is created when .save() is called for the first time.

        Parameters
        ----------
            data:
                the data to be saved
            kwargs:
                only there to allow subclasses to override the save() method
        """

        if not self._save_worker:
            self._save_worker = self.SaveWorker(self._data_manager, self._save_buffer)
            self._save_worker.start()
        self._save_buffer.put(data)

    def shutdown(self):
        """
        Waits for load and save worker to finish their tasks.
        Clears all the buffers, terminates all workers and prepares the buffered data manager to be used again.
        Should be called when one is done with iterating over the samples to allow the python process to end.
        """

        self._buffered_data_loader.shutdown()

        # Possibly awake blocking SaveWorker and signalize that no more data
        # will be put to the save buffer, i.e., the worker can shutdown
        self._save_buffer.put(_QUEUE_END_MSG)
        if self._save_worker:
            self._save_worker.join()

        self._save_buffer.queue.clear()
        self._save_worker = None

    class SaveWorker(Thread):
        """
        Background thread that waits for data to be saved on the internal save buffer.
        Will run until a special DONE MESSAGE is put onto the queue.
        """

        _data_manager: BaseDataManager
        _save_buffer: Queue

        def __init__(self, data_manager: BaseDataManager, save_buffer: Queue):
            Thread.__init__(self)
            self._data_manager = data_manager
            self._save_buffer = save_buffer

        def run(self) -> None:
            while True:
                data = self._save_buffer.get()
                if data == _QUEUE_END_MSG:
                    return
                with Timing() as t:
                    self._data_manager._save(data)

                try:
                    logging.info(f"Saving {len(data)} samples took {t[0]:0.3f} seconds")
                except TypeError:
                    # If data does not have length, don't log anything
                    pass

    def _save(self, data: Any):
        self.save(data)



import random

import numpy as np


def make_deterministic(seed: int = 0):
    try:
        import torch
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
    except ImportError:
        pass

    random.seed(seed)
    np.random.seed(seed)

    # torch.backends.cudnn.deterministic = True
    # torch.backends.cudnn.benchmark = False



import os
import re
from pathlib import Path
from shutil import copy2
from typing import Callable, Union

import PIL
from PIL import Image

from elias.util.range import IndexRangeBundle


def ensure_file_ending(file: str, ending: str) -> str:
    """
    Ensures that the given file path has the specified ending.

    Parameters
    ----------
        file: path to file
        ending: extension that the file should have

    Returns
    -------
        the file path with the specified extension if it did not already have that ending
    """

    return file if file.endswith(ending) else f"{file}.{ending}"


def ensure_directory_exists_for_file(path: Union[str, Path]):
    """
    Ensures that the folder to the specified path exists and creates a nested folder structure if necessary.
    Be careful to use trailing '/' if `path` directly constitutes the target folder (and not an arbitrary file
    within that folder).

    Parameters
    ----------
        path: path to the file or folder for which an underlying directory structure will be ensured
    """

    Path(os.path.dirname(str(path))).mkdir(parents=True, exist_ok=True)


def ensure_directory_exists(path: Union[str, Path]):
    """
    Ensures that the specified folder exists and creates a nested folder structure if necessary.

    Parameters
    ----------
        path: path to the folder which should exist
    """

    Path(str(path)).mkdir(parents=True, exist_ok=True)


def clear_directory(path: Union[str, Path]):
    """
    Clears all files from the specified directory, but keeps the directory itself.
    If the directory does not exist, nothing happens.
    """

    path = Path(str(path))
    if path.exists():
        for f in Path(path).iterdir():
            f.unlink()


# ==========================================================
# Copy utils
# ==========================================================

def _get_index_range_lambda(range_specifier: str) -> Callable[[int], bool]:
    if range_specifier == "":
        return lambda number: True
    else:
        index_range_bundle = IndexRangeBundle.from_description(range_specifier)
        return lambda number: number in index_range_bundle
        # range_specifier = range_specifier.replace("\\", "")  # Get rid of escape characters \\
        # index_ranges = [IndexRange.from_description(description) for description in
        #                 range_specifier.split(',')]
        # return lambda number: any([number in index_range for index_range in index_ranges])


def copy_transfer_fn(source_path: str, target_path: str):
    print(source_path, " -> ", target_path)
    ensure_directory_exists_for_file(target_path)
    copy2(source_path, target_path)


def copy_and_downsample_transfer_fn(source_path: str,
                                    target_path: str,
                                    downsample_factor: float):
    print(source_path, " -> ", target_path)
    ensure_directory_exists_for_file(target_path)

    image = Image.open(source_path)
    downsampled_size = (int(image.size[0] / downsample_factor),
                        int(image.size[1] / downsample_factor))
    image = image.resize(downsampled_size,
                         resample=PIL.Image.Resampling.LANCZOS)  # TODO: Is this the best resampling?
    image.save(target_path)


def smart_copy(
        source_files_format: str,
        target_files_format: str,
        transfer_fn: Callable[[str, str], None] = copy_transfer_fn):
    """
    Copies all the files that match the specified `source_files_format` and copies + renames them according to
    the wildcards used in `target_files_format`.

    Parameters
    ----------
        source_files_format:
            wildcarded path to source files, .e.g., C:/CaptureData/005/{s:}/cam_{c:}/image_{t:}.png
        target_files_format:
            wildcarded path to destination where files should be copied.
            All wildcards used in `source_files_format` must be reused.
            E.g., H:/CaptureData/005/sequence_{s:}_image_{t:}_cam_{c:}.png
        transfer_fn:
            A function that is executed for each pair of source/target paths.
            The default transfer_fn simply copies the source to the target file.
            Other transfer functions such as move instead of copy or copy + downsample are possible as well

    """

    # Traverse source file path up until the path does not contain any wildcards anymore.
    # From here, we can list all sub files and will find all relevant matches
    root_folder = Path(source_files_format)
    contains_wildcard_pattern = re.compile(r"{.*?:.*?}")
    while contains_wildcard_pattern.search(str(root_folder)):
        root_folder = root_folder.parent

    # Format now only contains parts with wildcards
    source_files_format = os.path.relpath(source_files_format, root_folder)

    print("Detected Root Folder:", root_folder)
    print("Copy files that match:", source_files_format)

    # Translate specified f-string into a regex
    # {var:} -> ((?P<var>.*)
    # This is necessary, as we cannot match f-strings against file names, but with regexes we can
    source_files_format = re.escape(source_files_format)

    decimal_wildcard_pattern = re.compile(r"\\{([^{}]*?):([^{}]*?)d\\}")
    generic_wildcard_pattern = re.compile(r"\\{([^{}]]*?):[^{}]*?\\}")
    decimal_wildcards = {wildcard: range_specifier
                         for wildcard, range_specifier
                         in decimal_wildcard_pattern.findall(source_files_format)}
    for wildcard, range_specifier in decimal_wildcards.items():
        decimal_wildcards[wildcard] = _get_index_range_lambda(range_specifier)

    source_files_regex = re.sub(decimal_wildcard_pattern, r"(?P<\g<1>>\\d*?)", source_files_format)
    source_files_regex = re.sub(generic_wildcard_pattern, r"(?P<\g<1>>.*?)", source_files_regex)
    print("Regex: ", source_files_regex)
    source_files_regex = re.compile(source_files_regex)

    for path in root_folder.rglob("*"):
        relevant_path = os.path.relpath(path, root_folder)
        match = re.match(source_files_regex, relevant_path)
        if match:
            match_arguments = match.groupdict()
            skip = False
            for key, value in match_arguments.items():
                if key in decimal_wildcards:
                    value = int(value)
                    include_value = decimal_wildcards[key]
                    if not include_value(value):
                        skip = True

                    match_arguments[key] = int(value)

            if not skip:
                target_path = target_files_format.format(**match_arguments)
                transfer_fn(path, target_path)



from shutil import rmtree
from typing import TypeVar, Type, Union

from elias.folder.run import RunFolder
from elias.manager.analysis import Analysis

_AnalysisType = TypeVar("_AnalysisType", bound=Analysis)


class AnalysisFolder(RunFolder[_AnalysisType]):

    _cls_run_manager: Type[_AnalysisType]

    def __init__(self, analysis_folder: str, name_format: str = '$_*', localize_via_run_name: bool = False):
        super(AnalysisFolder, self).__init__(analysis_folder, name_format, localize_via_run_name=localize_via_run_name)

    def open_analysis(self, analysis_name_or_id: Union[str, int]) -> _AnalysisType:
        analysis_name = self.resolve_run_name(analysis_name_or_id)
        return self._cls_run_manager.from_location(self._folder.get_location(),
                                                   analysis_name,
                                                   localize_via_analysis_name=self._localize_via_run_name)

    def new_analysis(self, analysis_name: str) -> _AnalysisType:
        new_analysis_name = self.generate_run_name(name=analysis_name)
        return self.open_analysis(new_analysis_name)

    def delete_analysis(self, analysis_name_or_id: Union[str, int]):
        analysis_name = self.resolve_run_name(analysis_name_or_id)
        rmtree(f"{self._folder.get_location()}/{analysis_name}")


import re
from abc import abstractmethod
from typing import Union, List

_RANGE_EXTRACTOR = re.compile(r"(-?\d+)-(-?\d+)")
_STEP_RANGE_EXTRACTOR = re.compile(r"(\d+)n(\+(\d+))?")


class ImplementsContains:

    @abstractmethod
    def __contains__(self, idx: int) -> bool:
        pass


class StepRange(ImplementsContains):
    step_size: int
    offset: int = 0

    def __init__(self, step_size: int, offset: int = 0):
        self.step_size = step_size
        self.offset = offset

    def __contains__(self, idx: int) -> bool:
        return (idx - self.offset) % self.step_size == 0


class IndexRange:
    """
    Represents an interval that is closed at both ends: [a, b].
    Useful for parsing user-defined ranges from the command line, e.g., 10-25.
    Negative numbers can be passed as well, in which case it is necessary to call `resolve()` with the actual number
    of items that the indices within the IndexRange refer to.
    E.g., 5--1 will resolve to 5-19 for a sequence of 20 items, or 5-99 for a sequence of 100 items.

    Provides parsing intervals from textual range specifications:
     - 1-5 -> [1, 5]
     - 1 -> [1, 1]
     - 7--1 -> resolve(20) -> [7, 19]
    """

    start_idx: int
    end_idx: int
    resolved: bool

    def __init__(self, start_idx: int, end_idx: int):
        assert start_idx <= 0 <= end_idx \
               or end_idx <= 0 <= start_idx \
               or start_idx <= end_idx <= 0 \
               or 0 <= start_idx <= end_idx, \
            "start_idx has to be smaller than end_idx"

        self.start_idx = start_idx
        self.end_idx = end_idx
        self.resolved = start_idx >= 0 and end_idx >= 0

    @staticmethod
    def from_description(index_range_description: str) -> Union['IndexRange', StepRange]:
        try:
            # Just a single number n. This is ok, will lead to an interval [n, n] that only accepts that number
            number = int(index_range_description)
            return IndexRange(number, number)
        except ValueError:
            pass

        step_range = _STEP_RANGE_EXTRACTOR.match(index_range_description)
        if step_range:
            step_size = int(step_range.group(1))  # The '5' of 5n+3
            offset = step_range.group(3)  # The '3' of 5n+3
            if offset is None:
                offset = 0
            else:
                offset = int(offset)
            assert 0 <= offset < step_size, f"offset should be between 0 and step_size {step_size}. Got {offset}"
            return StepRange(step_size, offset)

        frame_range = _RANGE_EXTRACTOR.search(index_range_description)
        # Ensure to convert negative indices to actual list indices
        start_idx = int(frame_range.group(1))
        end_idx = int(frame_range.group(2))

        return IndexRange(start_idx, end_idx)

    def resolve(self, n_samples: int) -> 'IndexRange':
        assert n_samples > abs(self.end_idx) and n_samples > abs(self.start_idx), \
            f"n_samples of {n_samples} is too small for index range [{self.start_idx}: {self.end_idx}]"

        self.start_idx = self.start_idx % n_samples
        self.end_idx = self.end_idx % n_samples
        self.resolved = True

        return self

    def __contains__(self, idx: int) -> bool:
        assert self.resolved, "Can only use 'in' on resolved index ranges"
        return self.start_idx <= idx <= self.end_idx

    def __len__(self) -> int:
        assert self.resolved, "Can only compute length on resolved index ranges"

        return self.end_idx - self.start_idx + 1  # + 1 because right end of interval is inclusive

    def __iter__(self):
        assert self.resolved, "Cannot iterate an unresolved index range"

        return iter(range(self.start_idx, self.end_idx + 1))


class IndexRangeBundle:

    def __init__(self, index_ranges: List[IndexRange]):
        self._index_ranges = index_ranges

    @staticmethod
    def from_description(range_specifier: str) -> 'IndexRangeBundle':
        range_specifier = range_specifier.replace("\\", "")  # Get rid of escape characters \\
        index_ranges = [IndexRange.from_description(description) for description in
                        range_specifier.split(',')]
        return IndexRangeBundle(index_ranges)

    def resolve(self, n_samples: int):
        for index_range in self._index_ranges:
            index_range.resolve(n_samples)

    def __contains__(self, idx: int) -> bool:
        for index_range in self._index_ranges:
            if idx in index_range:
                return True

        return False



import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Type, TypeVar, Generic, Optional, List, Union

from silberstral import reveal_type_var

from elias.manager.artifact import ArtifactManager, ArtifactType
from elias.config import Config
from elias.folder.folder import Folder

_ModelConfigType = TypeVar('_ModelConfigType', bound=Config)
_OptimizationConfigType = TypeVar('_OptimizationConfigType', bound=Config)
_DatasetConfigType = TypeVar('_DatasetConfigType', bound=Config)
_TrainSetupType = TypeVar('_TrainSetupType', bound=Config)
_EvaluationResultType = TypeVar('_EvaluationResultType', bound=Config)
_EvaluationConfigType = TypeVar('_EvaluationConfigType', bound=Config)
_ModelType = TypeVar('_ModelType')


class ModelManager(ABC,
                   Generic[
                       _ModelType, _ModelConfigType, _OptimizationConfigType, _DatasetConfigType, _TrainSetupType,
                       _EvaluationConfigType, _EvaluationResultType],
                   ArtifactManager):
    """
    Type Vars
    ---------

        _ModelType:
            The type of the model being handled by this model manager. Only there for type hinting, can be None
        _ModelConfigType:
            The class of the model configuration (stored in ``model_config.json``) which is assumed to be a dataclass
            subclassing :class:`elias.config.Config`. :meth:`save_model_config` and :meth:`load_model_config`
            take/retrieve the dataset configuration as a Python object of this class.
            Typical contents of a model_config:
                * Model Architecture
                * Model Type
                * Layers
                * Anything that affects learnable weights
                * Input/Output format
        _OptimizationConfigType:
            The class of the optimization configuration (stored in ``optimization_config.json``).
            Works analog to _ModelConfigType
            Typical contents of an optimization_config:
                * Loss
                * Regularization
                * Type of Optimizer (Adam, RMSProp, ...)
                * Learning Rate scheduling
        _DatasetConfigType:
            The class of the dataset configuration (stored in ``dataset_config.json``).
            Works analog to _ModelConfigType
            Typical contents of a dataset_config:
                * Dataset Type
                * Dataset Version
                * Partition
                * Image Augmentation
                * Any online Data Preprocessing
                * Anything that affects the format of batches fed to the model
        _TrainSetupType:
            The class of the train setup (stored in ``train_setup.json``).
            Works analog to _ModelConfigType
            Typical contents of a train_setup:
                * Seed
                * Batch size
                * train/valid split
                * Logging/Monitoring config
                * Multithreading
                * Gradient Accumulation
        _EvaluationConfigType:
            The class of the evaluation config.
            The file name depends on `evaluation_name_format` and `artifact_type`.
            E.g., if `evaluation_name_format="evaluation_epoch_$"` and JSON artifact the file name may be
            "evaluation_epoch_100_config.json".
            Can be None, if evaluation configs shall not be persisted
            Typical contents of an evaluation config:
                * Which metrics to evaluation
        _EvaluationResultType:
            The class of the evaluation result.
            The file name depends on `evaluation_name_format` and `artifact_type`.
            E.g., if `evaluation_name_format="evaluation_epoch_$"` and JSON artifact the file name may be
            "evaluation_epoch_100.json"
            Typical contents of an evaluation result:
                * Scores
    """

    _folder: Folder
    _run_name: str
    _checkpoint_name_format: Optional[str]
    _evaluation_name_format: Optional[str]
    _evaluation_config_name_format: Optional[str]
    _cls_model_config: Type[_ModelConfigType]
    _cls_optimization_config: Type[_OptimizationConfigType]
    _cls_dataset_config: Type[_DatasetConfigType]
    _cls_train_setup: Type[_TrainSetupType]
    _cls_evaluation_result: Type[_EvaluationResultType]
    _cls_evaluation_config: Type[_EvaluationConfigType]

    def __init__(self,
                 model_store_path: str,
                 run_name: str,
                 checkpoint_name_format: Optional[str] = None,
                 checkpoints_sub_folder: Optional[str] = None,
                 evaluation_name_format: Optional[str] = None,
                 artifact_type: ArtifactType = ArtifactType.JSON):
        """
        Creates a model manager that provides utilities for loading/storing checkpoints, configs and evaluations
        for a model run.

        Parameters
        ----------
            model_store_path:
                Root folder of runs
            run_name:
                Name of this specific run
            checkpoint_name_format:
                If storing/loading checkpoints via ID, e.g., epoch, is desired, a file name format has to be specified.
                It has to contain exactly one `$` indicating the checkpoint ID, e.g., `checkpoint-$.ckpt`
            checkpoints_sub_folder:
                In case the checkpoints are not stored in the main model folder, one can specify a sub folder within
                the main folder that will hold checkpoint files
            evaluation_name_format:
                If storing/loading evaluations via checkpoint ID is desired, a file name format has to be specified.
                It has to contain exactly one `$` indicating the checkpoint ID and EXCLUDING the file name suffix.
                This will be inferred automatically from the specified `artifact_type`.
                Example for `evaluation_name_format`: `evaluation_$`
            artifact_type:
                In which format, e.g., JSON or YAML, configs and evaluations should be stored
        """

        assert Path(
            f"{model_store_path}/{run_name}").is_dir(), f"Could not find directory '{model_store_path}/{run_name}'. Is the run name {run_name} correct?"
        super(ModelManager, self).__init__(f"{model_store_path}/{run_name}", artifact_type=artifact_type)

        self._folder = Folder(f"{model_store_path}/{run_name}")
        if checkpoints_sub_folder is None:
            self._checkpoints_folder = self._folder
        else:
            self._checkpoints_folder = Folder(f"{model_store_path}/{run_name}/{checkpoints_sub_folder}")

        self._run_name = run_name
        self._checkpoint_name_format = checkpoint_name_format

        if evaluation_name_format is not None:
            # Append correct artifact type suffix, e.g., .json or .yaml
            self._evaluation_name_format = f"{evaluation_name_format}.{self._artifact_type.get_file_ending()}"
            # Evaluation configs are hard-coded to have "_config" preceding the suffix
            self._evaluation_config_name_format = \
                f"{evaluation_name_format}_config.{self._artifact_type.get_file_ending()}"
        else:
            self._evaluation_name_format = None
            self._evaluation_config_name_format = None

        self._cls_model_config = reveal_type_var(self, _ModelConfigType)
        self._cls_optimization_config = reveal_type_var(self, _OptimizationConfigType)
        self._cls_dataset_config = reveal_type_var(self, _DatasetConfigType)
        self._cls_train_setup = reveal_type_var(self, _TrainSetupType)
        self._cls_evaluation_result = reveal_type_var(self, _EvaluationResultType)
        self._cls_evaluation_config = reveal_type_var(self, _EvaluationConfigType)

    @classmethod
    def from_location(cls: Type['ModelManager'],
                      model_store_path: str,
                      run_name: str,
                      localize_via_run_name: bool = False) -> 'ModelManager':
        """
        Creates a model manager for the specified location with default parameters.
        If the subclass constructor takes different arguments than the location, this needs to be overridden
        by subclasses to ensure that there is a instantiation method that takes exactly one argument.

        Parameters
        ----------
            model_store_path:
                path to the folder containing model runs
            run_name:
                name of the run
            localize_via_run_name:
                whether only the run name should be used to find model folder

        Returns
        -------
            a model manager for the specified run
        """

        try:
            if localize_via_run_name:
                new_model_manager = cls(run_name)
            else:
                new_model_manager = cls(model_store_path, run_name)
        except TypeError as e:
            raise NotImplementedError(f"Could not construct model manager {cls} with location and run name parameter. "
                                      f"Please override from_location() to match the class __init__() method"
                                      f"Other possible error: {e}")
        return new_model_manager

    # -------------------------------------------------------------------------
    # Building/Storing/Loading Models
    # -------------------------------------------------------------------------

    def get_checkpoint_path(self, checkpoint_name_or_id: Union[str, int]):
        assert self._checkpoint_name_format is not None, "Cannot get checkpoint path, no file name format specified"

        checkpoint_id = self._resolve_checkpoint_id(checkpoint_name_or_id)
        self._checkpoints_folder.substitute(self._checkpoint_name_format, checkpoint_id)
        checkpoint_file_name = self._checkpoints_folder.get_file_name_by_numbering(self._checkpoint_name_format,
                                                                                   checkpoint_id)

        checkpoint_path = f"{self._checkpoints_folder.get_location()}/{checkpoint_file_name}"
        return checkpoint_path

    def get_checkpoint_folder(self) -> str:
        return self._checkpoints_folder.get_location()

    def list_checkpoints(self) -> List[str]:
        assert self._checkpoint_name_format is not None, "Cannot list checkpoints, no file name format specified"

        checkpoint_ids_and_names = self._checkpoints_folder.list_file_numbering(self._checkpoint_name_format)
        if len(checkpoint_ids_and_names) == 0:
            return []

        checkpoint_ids, checkpoint_names = zip(*checkpoint_ids_and_names)
        last_neg_id = None
        for idx, checkpoint_id in enumerate(checkpoint_ids):
            if checkpoint_id < 0:
                last_neg_id = idx

        checkpoint_names = list(checkpoint_names)
        if last_neg_id is not None:
            checkpoint_names = checkpoint_names[last_neg_id + 1:] + checkpoint_names[:last_neg_id + 1]

        return checkpoint_names

    def list_checkpoint_ids(self) -> List[int]:
        assert self._checkpoint_name_format is not None, "Cannot list checkpoints, no file name format specified"

        checkpoint_ids = self._checkpoints_folder.list_file_numbering(self._checkpoint_name_format,
                                                                      return_only_numbering=True)
        last_neg_id = None
        for idx, checkpoint_id in enumerate(checkpoint_ids):
            if checkpoint_id < 0:
                last_neg_id = idx

        if last_neg_id is not None:
            checkpoint_ids = checkpoint_ids[last_neg_id + 1:] + checkpoint_ids[:last_neg_id + 1]

        return checkpoint_ids

    def store_checkpoint(self, model: _ModelType, checkpoint_name_or_id: Union[str, int], **kwargs):
        if isinstance(checkpoint_name_or_id, int):
            assert self._checkpoint_name_format is not None, \
                f"Cannot store checkpoint with id {checkpoint_name_or_id} since no checkpoint name format was specified"
            checkpoint_file_name = self._checkpoints_folder.substitute(self._checkpoint_name_format,
                                                                       checkpoint_name_or_id)
        else:
            checkpoint_file_name = checkpoint_name_or_id

        self._store_checkpoint(model, checkpoint_file_name, **kwargs)

    def load_checkpoint(self, checkpoint_name_or_id: Union[str, int], **kwargs) -> _ModelType:
        checkpoint_id = self._resolve_checkpoint_id(checkpoint_name_or_id)
        checkpoint_file_name = self._checkpoints_folder.get_file_name_by_numbering(self._checkpoint_name_format,
                                                                                   checkpoint_id)

        return self._load_checkpoint(checkpoint_file_name, **kwargs)

    def delete_checkpoint(self, checkpoint_name_or_id: Union[str, int]):
        checkpoint_id = self._resolve_checkpoint_id(checkpoint_name_or_id)
        checkpoint_file_name = self._checkpoints_folder.get_file_name_by_numbering(self._checkpoint_name_format,
                                                                                   checkpoint_id)

        os.remove(f"{self._checkpoints_folder}/{checkpoint_file_name}")

    def build_model(self,
                    model_config: Optional[_ModelConfigType] = None,
                    optimization_config: Optional[_OptimizationConfigType] = None,
                    load_optimization_config: bool = False,
                    **kwargs) -> _ModelType:
        model_config = self.load_model_config() if model_config is None else model_config
        optimization_config = self.load_optimization_config() \
            if optimization_config is None and load_optimization_config \
            else optimization_config

        return self._build_model(model_config, optimization_config, **kwargs)

    @abstractmethod
    def _build_model(self,
                     model_config: _ModelConfigType,
                     optimization_config: Optional[_OptimizationConfigType] = None,
                     **kwargs) -> _ModelType:
        """
        Should be overwritten by subclass to construct an actual model object given the model and optimization configs

        Parameters
        ----------
            model_config:
                the configuration of the model to construct
            optimization_config:
                specifies what kind of optimization the model should be prepared for

        Returns
        -------
            A fully operable model with the given configs. No trained parameters have been applied yet
        """

        pass

    @abstractmethod
    def _store_checkpoint(self, model: _ModelType, checkpoint_file_name: str, **kwargs):
        """
        Should be overwritten by subclass to specify how exactly the model will be persisted

        Parameters
        ----------
            model:
                The model to persist
            checkpoint_file_name:
                The full name for the checkpoint excluding the path to the folder
        """

        pass

    @abstractmethod
    def _load_checkpoint(self, checkpoint_file_name: Union[str, int], **kwargs) -> _ModelType:
        """
        Should be overwritten by subclass to specify how exactly a persisted checkpoint should be loaded.
        Typically, this method will make use of `_build_model()` and then apply the loaded trained weights from the
        checkpoint file.

        Parameters
        ----------
            checkpoint_file_name:
                name of the file containing learned weights

        Returns
        -------
            A fully operable model with the learned weights from the specified checkpoint file
        """

        pass

    # -------------------------------------------------------------------------
    # Storing/Loading Configs
    # -------------------------------------------------------------------------

    def store_model_config(self, model_config: _ModelConfigType):
        self._save_artifact(model_config.to_json(), "model_config")

    def load_model_config(self) -> _ModelConfigType:
        return self._load_config(self._cls_model_config, "model_config")

    def store_optimization_config(self, optimization_config: _OptimizationConfigType):
        self._save_artifact(optimization_config.to_json(), "optimization_config")

    def load_optimization_config(self) -> _OptimizationConfigType:
        return self._load_config(self._cls_optimization_config, "optimization_config")

    def store_dataset_config(self, dataset_config: _DatasetConfigType):
        self._save_artifact(dataset_config.to_json(), "dataset_config")

    def load_dataset_config(self) -> _DatasetConfigType:
        return self._load_config(self._cls_dataset_config, "dataset_config")

    def store_train_setup(self, train_setup: _TrainSetupType):
        self._save_artifact(train_setup.to_json(), "train_setup")

    def load_train_setup(self) -> _TrainSetupType:
        return self._load_config(self._cls_train_setup, "train_setup")

    # -------------------------------------------------------------------------
    # Evaluations
    # -------------------------------------------------------------------------

    def list_evaluations(self) -> List[str]:
        assert self._evaluation_name_format is not None, "Cannot list evaluations, no file name format specified"
        evaluation_files = self._folder.list_file_numbering(self._evaluation_name_format, return_only_file_names=True)

        return evaluation_files

    def list_evaluated_checkpoint_ids(self) -> List[int]:
        assert self._evaluation_name_format is not None, "Cannot list evaluations, no file name format specified"
        checkpoint_ids = self._folder.list_file_numbering(self._evaluation_name_format, return_only_numbering=True)

        return checkpoint_ids

    def store_evaluation_result(self, evaluation_result: _EvaluationResultType, checkpoint_name_or_id: Union[str, int]):
        if isinstance(checkpoint_name_or_id, str) and self._evaluation_name_format is None:
            # Assume that given checkpoint name should be file name for evaluation
            self._save_artifact(evaluation_result.to_json(), checkpoint_name_or_id)
        else:
            assert self._evaluation_name_format is not None, "Cannot store evaluation, no file name format specified"
            checkpoint_id = self._resolve_checkpoint_id(checkpoint_name_or_id)
            evaluation_file_name = self._folder.substitute(self._evaluation_name_format, checkpoint_id)
            self._save_artifact(evaluation_result.to_json(), evaluation_file_name)

    def load_evaluation_result(self, checkpoint_name_or_id: Union[str, int]) -> _EvaluationResultType:
        if isinstance(checkpoint_name_or_id, str) and self._evaluation_name_format is None:
            # Assume that given checkpoint name should be file name for evaluation
            return self._load_config(self._cls_evaluation_result, checkpoint_name_or_id)
        else:
            assert self._evaluation_name_format is not None, "Cannot load evaluation, no file name format specified"

            checkpoint_id = self._resolve_checkpoint_id(checkpoint_name_or_id)
            evaluation_file_name = self._folder.substitute(self._evaluation_name_format, checkpoint_id)
            return self._load_config(self._cls_evaluation_result, evaluation_file_name)

    def store_evaluation_config(self, evaluation_config: _EvaluationConfigType, checkpoint_name_or_id: Union[str, int]):
        assert self._evaluation_config_name_format is not None, \
            "Cannot store evaluation config, no file name format specified"
        checkpoint_id = self._resolve_checkpoint_id(checkpoint_name_or_id)
        evaluation_file_name = self._folder.substitute(self._evaluation_config_name_format, checkpoint_id)
        self._save_artifact(evaluation_config.to_json(), evaluation_file_name)

    def load_evaluation_config(self, checkpoint_name_or_id: Union[str, int]):
        assert self._evaluation_config_name_format is not None, \
            "Cannot load evaluation config, no file name format specified"

        checkpoint_id = self._resolve_checkpoint_id(checkpoint_name_or_id)
        evaluation_file_name = self._folder.substitute(self._evaluation_config_name_format, checkpoint_id)
        return self._load_config(self._cls_evaluation_config, evaluation_file_name)

    # -------------------------------------------------------------------------
    # Getters
    # -------------------------------------------------------------------------

    def get_run_name(self) -> str:
        return self._run_name

    def get_location(self) -> str:
        return self._folder.get_location()

    def get_model_store_path(self) -> str:
        return self._folder.get_location()

    # -------------------------------------------------------------------------
    # Helper Methods
    # -------------------------------------------------------------------------

    def _load_config(self, artifact_cls: Type[Config], artifact_name: str) -> Config:
        if isinstance(None, artifact_cls):
            raise NotImplementedError(f"Cannot load `{artifact_name}` as its corresponding type is not defined")
        return artifact_cls.from_json(self._load_artifact(artifact_name))

    def _resolve_checkpoint_id(self, checkpoint_name_or_id: Union[str, int]) -> int:
        # Special handling for allowing 'latest' and 'last' as checkpoint names
        if checkpoint_name_or_id in {'latest', 'last'}:
            checkpoint_name_or_id = -1

        if isinstance(checkpoint_name_or_id, int):
            assert self._checkpoint_name_format is not None, \
                "Cannot resolve checkpoint by id, no file name format specified"
            if checkpoint_name_or_id < 0:
                # Find latest checkpoint
                assert checkpoint_name_or_id == -1, \
                    f"Only -1 is allowed as negative checkpoint id, got `{checkpoint_name_or_id}`"

                checkpoint_ids, checkpoint_names = zip(
                    *self._checkpoints_folder.list_file_numbering(self._checkpoint_name_format))
                if -1 in checkpoint_ids:
                    checkpoint_id = -1
                else:
                    # No checkpoint with id -1 present => Take checkpoint with largest ID instead
                    assert len(checkpoint_ids) > 0, \
                        f"Cannot find latest checkpoint, no checkpoints found in {self._checkpoints_folder.get_location()}"
                    checkpoint_id = checkpoint_ids[-1]
            else:
                # Find corresponding name for checkpoint id
                checkpoint_id = checkpoint_name_or_id

            return checkpoint_id
        else:
            checkpoint_id = self._checkpoints_folder.get_numbering_by_file_name(self._checkpoint_name_format,
                                                                                checkpoint_name_or_id)
            return checkpoint_id
